#include "Mat4f.h"

// ====== CONSTRUCTORS (MOVE/COPY) / DESTRUCTORS ============================//
Mat4f::Mat4f()
{
	Mat4fHandle tmp ( new ARRAY_16f );
	m_ptr = std::move( tmp );	//take ownership
}

Mat4f::Mat4f( std::initializer_list< float > list )
{
	if( list.size() != MAT4F_NUM_ELEM )
		throw std::out_of_range("Mat4f init list size not 16");

	Mat4fHandle tmp ( new ARRAY_16f );
	std::copy_n( 	list.begin(), 		//source
			MAT4F_NUM_ELEM, 	//number of copies
			tmp->begin() );		//destination

	m_ptr = std::move( tmp );		//take ownership
}

Mat4f::Mat4f( Mat4f && moved )
{
	m_ptr = std::move( moved.m_ptr );	//take ownership
}

Mat4f::Mat4f( const Mat4f & copied )
{
	Mat4fHandle tmp ( new ARRAY_16f );
	std::copy_n( 	copied.begin(), 	//source
			MAT4F_NUM_ELEM, 	//number of copies
			tmp->begin() );		//destination 

	m_ptr = std::move( tmp ); 		//take ownership
}

Mat4f::~Mat4f()
{}
// ==========================================================================//

// =========== OPERATORS ====================================================//

Mat4f & Mat4f::operator=( const Mat4f & copied )
{
	if( this != &copied )
	{
		Mat4f tmp( copied );
		*this = std::move( tmp );
	}
	return *this;
}

Mat4f & Mat4f::operator=( Mat4f && moved )
{
	m_ptr = std::move( moved.m_ptr );
	return *this;
}


// Cause we store array in row major form, the elements are laid out row, after row.
// So to jump to the right row in the array, row*4, then plus the column to get the
// right matrix row,column element.
float & Mat4f::operator()( int row, int column )
{
	if( !isValidDimIndex( row ) || !isValidDimIndex( column ) )
		throw std::out_of_range("Mat4f (row, column) out of range (0-3,0-3).");
	return m_ptr->at( row*MAT4F_DIM + column ); 
}

float Mat4f::operator()( int row, int column ) const
{
	if( !isValidDimIndex( row ) || !isValidDimIndex( column ) )
		throw std::out_of_range("Mat4f (row, column) out of range (0-3,0-3).");
	return m_ptr->at( row*MAT4F_DIM + column );
}

float & Mat4f::operator[]( int element )
{
	if( !isValidElementIndex( element ) )
		throw std::out_of_range("Mat4f (element) out of range (0-15).");
	return m_ptr->at( element );
}

float Mat4f::operator[]( int element ) const
{
	if( !isValidElementIndex( element ) )
		throw std::out_of_range("Mat4f (element) out of range (0-15).");
	return m_ptr->at( element );
}

/*
Mat4f Mat4f::operator+( Mat4f other ) const
{
	// TODO
}

Mat4f Mat4f::operator*( const Mat4f & other ) const
{
	// TODO
}

Mat4f Mat4f::operator*( float scalar ) const
{
	// TODO
}
*/

// ==========================================================================//


// Just helpful stuff...
// Play around with it, google or ask me :)
float const * Mat4f::data() const
{
	return m_ptr->data();
}

Mat4f::ARRAY_16f::iterator Mat4f::begin()
{
	return m_ptr->begin();
}

Mat4f::ARRAY_16f::iterator Mat4f::end()
{
	return m_ptr->end();
}

Mat4f::ARRAY_16f::const_iterator Mat4f::begin() const
{
	return m_ptr->begin();
}

Mat4f::ARRAY_16f::const_iterator Mat4f::end() const
{
	return m_ptr->end();
}

bool Mat4f::isValidDimIndex( int idx ) const
{
	return idx >= 0 && idx < MAT4F_DIM;
}
bool Mat4f::isValidElementIndex( int idx ) const
{
	return idx >= 0 && idx < MAT4F_NUM_ELEM;
}

std::ostream & operator<<( std::ostream & out, const Mat4f & mat )
{
	std::ostream_iterator<int> out_it (out," ");
  	std::copy ( mat.begin(), mat.end(), out_it );
	return out;
}



